#include <assert.h>
#include <stdlib.h>

// Main file for CPU code.

#include "vmc.h"

#define __constant__
#include "cudaMain.cuh"

#define clock64 time_wall_int64
static inline void __syncthreads(){}

static inline int32_t
curand(random_data *buf)
{
  int32_t r;
  random_r(buf,&r);
  return r;
}

static inline float 
curand_uniform(random_data *buf)
{
  int32_t r;
  random_r(buf,&r);
  return (1.0f/(RAND_MAX)) * r;
}

const struct { int x; } threadIdx = { 0 };
const int block_size = 1;
const int warp_count = 1;

const int nneibs = 12;

#undef USECONST
#ifdef USECONST
const int npairs = 131;
const int nsites = 290;
const int psiinv_stride = 136;
#else
static int npairs;
static int nsites;
static int psiinv_stride;
#endif

typedef unsigned short Site_Idx;

class Computation_Thread {
public:
  Computation_Thread(int mid)
    :MID(mid),d_sqlat(c_sqlat),pairfunc(c_pairfunc),d_config(c_config)
  { cpu_sysAve(); }
  void cpu_sysAve();
  double cpu_EnergyOfCurrentConfiguration();
  void s_pairfunc_flatten(Pairfunction pairfunc, int newsite, int spin);
  void s_tpairfunc_flatten(Pairfunction pairfunc, int newsite, int spin);
  double cpu_DetPByDet
  (int ipair, int spin, int newsite,
   Pairfunction pairfunc, Configuration d_config);
  void cpu_UpdateConfig
  (int ipair, int ispin, int newsite, Pairfunction pairfunc, 
   double dpbdt, Configuration d_config);
  double cpu_d_UpdateConfig_d
  ( int ipair, int spin, int newsite, int jpair, int isite);

  const int MID;
  Lattice& d_sqlat;
  Pairfunction& pairfunc;
  Configuration& d_config;
  struct { int x; } blockIdx;

  // For entire rows or columns of phi
  double *s_psi_cache;

  clock_t time_update_config[2], time_detpbydet[2];
  clock_t time_d_update_d[2], time_d_update_dw[2];
  int time_update_config_cnt, time_detpbydet_cnt;
  int time_d_update_d_cnt, time_d_update_dw_cnt;
  int time_energy_exchange_term_cnt;
  clock_t timer_start_energy_exchange, time_energy_exchange;
};

void
cpu_set_vars
(montecarlo& mp, sqlat &latobj, phyparams& phyobj, int psiinv_stridep)
{
  assert( nneibs == latobj.nneibs );
#ifdef USECONST
  assert( npairs == mp.current.npairs );
  assert( nsites == mp.current.nsites );
  assert( psiinv_stride == psiinv_stridep );
#else
  npairs = mp.current.npairs;
  nsites = mp.current.nsites;
  psiinv_stride = psiinv_stridep;
#endif
}


#if 0
__shared__ Site_Idx s_plist[2][npairs];
__shared__ char s_latocc[nsites];
__shared__ Site_Idx s_whichpair[2*nsites];
__shared__ Site_Idx s_neiblist[nneibs*nsites];
__shared__ Site_Idx s_nearnp[4 * nsites];
#define Nearnp(idx) s_nearnp[idx]
#define PList(pair,spin) s_plist[spin][pair]
#define dlatocc(site) s_latocc[site]
#define Whichpair(site,spin) s_whichpair[(site)*2+(spin)]
#define Neiblist(site,jn) s_neiblist[nneibs*(site)+(jn)]
#else

#define Nearnp(idx) c_sqlat.nearnp[idx]
#define PList(pair,spin) c_config.plist[blockIdx.x*2*npairs+(pair)*2+(spin)]
#define dlatocc(site) c_config.latocc[blockIdx.x*nsites+(site)]
#define Whichpair(site,spin) \
  c_config.whichpair[blockIdx.x*2*nsites+(site)*2+(spin)]
#define Neiblist(site,jn) c_sqlat.neiblist[(jn)+nneibs*(site)]

#endif


#define PsiInv(r,c) \
  c_config.p_PsiInv[blockIdx.x*npairs*psiinv_stride+psiinv_stride*(r)+(c)]

#define Pairfunc(r,c) c_pairfunc.pairfn[(r)*nsites+(c)]
#define TPairfunc(r,c) c_pairfunc.Tpairfn[(c)*nsites+(r)]

#define PsiEltCol(c) s_psi_cache[c]
#define PsiEltRow(r) s_psi_cache[r]

#include "cpu_equil_accum.cpp"
