
#include "cpu_df.h"
#include <strings.h>

__attribute__((optimize("unroll-loops"))) double
Computation_Thread::cpu_EnergyOfCurrentConfiguration()
{
  double enekloc, otheloc, spkeloc[2];
  
  // ***---------------------------------------------------------    
  //        Kinetic energy part of the Hamiltonian
  //                          _ 
  //         H  = -\sum  t  ( c   c   )
  //          t    (i,j)  ij   is  js
  // ***---------------------------------------------------------    

  {
    const int ispin = 0;             // Down
    double sum = 0;

    for ( int jn = 0; jn < nneibs; jn++ )
      {
        double thd_sum_jn = 0;
        for ( int ipair = 0; ipair < npairs; ipair++ )
          {
            const int isite = PList(ipair,ispin);
            const int jsite = Neiblist(isite,jn);

            if ( dlatocc(jsite) == HL )
              {
                for ( int col = 0; col < npairs; col++ )
                  thd_sum_jn +=
                    TPairfunc(PList(col,1),jsite) * PsiInv(ipair,col);
              }
          }
	
        sum += thd_sum_jn * d_sqlat.thop_ishell[jn];
      }

    if ( !threadIdx.x )
      {
        enekloc = sum;
        spkeloc[ispin] = sum;
      }
  }
  {
    const int ispin = 1;  // Up
    const int thd_per_jn = 1;
    const int jn_per_block = 1;
    const int thd_per_ipair = 1;
    const int my_1st_ipair = threadIdx.x % thd_per_ipair;
    const int my_1st_row = threadIdx.x / thd_per_ipair % thd_per_jn;
    const int my_1st_jn = threadIdx.x / ( thd_per_ipair * thd_per_jn );
    double thread_sum = 0;
    
    for ( int jn = my_1st_jn; jn < nneibs; jn += jn_per_block )
      {
        double thd_sum_jn = 0;

        for (int ipair = my_1st_ipair; ipair < npairs; ipair += thd_per_ipair )
          {
            const int isite = PList(ipair,ispin);
            const int jsite = Neiblist(isite,jn);

            if ( dlatocc(jsite) == HL ) 
              for ( int row = my_1st_row; row < npairs; row += thd_per_jn )
                thd_sum_jn += 
                  Pairfunc(jsite,PList(row,0)) * PsiInv(row,ipair);
          }

        thread_sum += thd_sum_jn * d_sqlat.thop_ishell[jn];
      }

    const double sum = thread_sum;
    if ( !threadIdx.x )
      {
        enekloc += sum;
        spkeloc[ispin] = sum;
      }
  }  

	// ***---------------------------------------------------------
	//        Now compute the Exchange Term
	//                  _  _     1 
	//         H  = J ( S .S  - ---  n n  ) Term
	//          J        i  j    4    i j
	// ***---------------------------------------------------------
  if ( !threadIdx.x ) otheloc = 0.0;

# ifdef COLLECT_CYCLE_COUNTS
  if ( !threadIdx.x ) timer_start_energy_exchange = clock64(); 
# endif

  const int nearnsets = 2 * nsites;

  double ol_local = 0;

  const int tid = threadIdx.x;
  const int lane = tid & WP_MA;

#if 1

  int term_count_local = 0;

  // Compute each exchange term using all threads in block.
  for ( int inn = 0; inn < nearnsets; inn++ ) 
    {
      const int isite = d_sqlat.nearnp[inn*2+0];
      const int jsite = d_sqlat.nearnp[inn*2+1];
      // Note ispin is used as spin here.
      const int sispin = dlatocc(isite);
      const int jspin = dlatocc(jsite);
      
      if ( sispin + jspin == 1 ) 
	{
          const int jpair = Whichpair(jsite,jspin);
          const int sipair = Whichpair(isite,sispin);

          const double e = cpu_d_UpdateConfig_d
            ( sipair, sispin, jsite, jpair, isite);

	  if ( !threadIdx.x ) { ol_local -= e;  term_count_local++; }
	}
    }
#else

  const int wp_per_block = block_size >> WP_LG;
  float term_count_local = 0;
  const float term_local_inc = use_row_cache ? 1.0f / num_row_blocks : 1.0f;
  const int row_block_inc = use_row_cache ? row_block_size : npairs;

  for ( int start_row = 0; start_row < npairs; start_row += row_block_inc )
    {
      if ( use_row_cache )
        {
          __syncthreads();
          for ( int i=0; i<row_block_size; i++ )
            if ( tid < npairs && start_row + i < npairs )
              sb_psiinv_cache[i][tid] = PsiInv(start_row+i,tid);
          __syncthreads();
        }

      // Compute each exchange term with one warp.
      for ( int inn = wid; inn < nearnsets; inn += wp_per_block ) 
        {
          const int isite = Nearnp(inn*2+0);
          const int jsite = Nearnp(inn*2+1);
          // Note ispin is used as spin here.
          const int sispin = dlatocc(isite);
          const int jspin = dlatocc(jsite);

          if ( sispin + jspin == 1 ) 
            {
              const int jpair = Whichpair(jsite,jspin);
              const int sipair = Whichpair(isite,sispin);
              const bool want_up = use_row_cache ? true : sispin == UP;
              const int spin = want_up ? UP : DN;
              const bool fwd = want_up == ( sispin == UP );
              const int asite = fwd ? isite : jsite;
              const int bsite = fwd ? jsite : isite;
              const int apair = fwd ? sipair : jpair;
              const int bpair = fwd ? jpair : sipair;

              const double e = 
                use_row_cache
                ? d_UpdateConfig_d_warp_block
                (apair, bsite, pairfunc, d_config, bpair, asite, start_row)
                : d_UpdateConfig_d_warp
                (apair, spin, bsite, pairfunc, d_config, bpair, asite);

              term_count_local += term_local_inc;
              ol_local -= e;
            }
        }
    }

#endif


  if ( !lane )
    {
      ol_local -= term_count_local;
    }

  otheloc = ol_local;

# ifdef COLLECT_CYCLE_COUNTS
  time_energy_exchange_term_cnt += term_count_local;
# endif

  double energy = 0;
  if ( !threadIdx.x )
    {
      otheloc = Jij * otheloc / 2.0;
      energy = ( enekloc + otheloc ) / double( nsites );
      d_config.energy[MID] = energy;
      d_config.kinenergy[MID] = ( spkeloc[0] + spkeloc[1] ) / double( nsites );
      d_config.othenergy[MID] = otheloc;
      d_config.modified[MID] = false;
    }
  return energy;
}

extern "C" void*
cpu_sysAve(void *data)
{
  const int id = (size_t) data;
  Computation_Thread ct(id);
  return NULL;
}

void
Computation_Thread::cpu_sysAve()
{
  const int tid = 0;
  blockIdx.x = MID;

  MonteCarlo& d_mcarlo = c_mcarlo;
  double*& d_enearr = c_enearr;

  random_data localState;
  bzero(&localState,sizeof(localState));
  char rand_state[64];
  initstate_r(1 + MID,rand_state,sizeof(rand_state),&localState);

  s_psi_cache = new double[npairs];

  int idat, nbsamp;
  int ipair,ispin,oldsite,newsite;
  bool spin_flip;
  double eneloc;
  
# ifdef COLLECT_CYCLE_COUNTS
  clock_t timer_start_thread, timer_start_eoc, timer_start_findpairs;
  clock_t timer_start_move_eval, timer_start_update;
  clock_t time_move_eval, time_update, time_eoc, time_findpairs;
# endif

  if ( !tid )
    {
#     ifdef COLLECT_CYCLE_COUNTS
      timer_start_thread = clock64();
      time_eoc = time_findpairs = 0;
      time_energy_exchange_term_cnt = 0;
      time_energy_exchange = 0;
      time_detpbydet[0] = time_detpbydet[1] = 0;
      time_move_eval = time_update = 0;
      time_update_config[0] = time_update_config[1] = 0;
      time_update_config_cnt = time_detpbydet_cnt = 0;
      time_d_update_d[0] = time_d_update_d[1] = 0;
      time_d_update_dw[0] = time_d_update_dw[1] = 0;
      time_d_update_d_cnt = 0;
      time_d_update_dw_cnt = 0;
#     endif
      
      idat = 0; eneloc = 0.0; nbsamp = 0;
    }

  for ( int tid=0; tid<npairs; tid++ )
    {
      PList(tid,0) = d_config.plist[MID*2*npairs+(tid)*2];
      PList(tid,1) = d_config.plist[MID*2*npairs+(tid)*2+1];
      for ( int row=0; row<npairs; row++ )
        PsiInv(row,tid) = d_config.PsiInv[MID*npairs*npairs+npairs*row+tid];
    }

  for ( int site = threadIdx.x; site < nsites; site += block_size )
    {
      dlatocc(site) = d_config.latocc[MID*nsites+(site)];
      Whichpair(site,0) = d_config.whichpair[MID*2*nsites+site*2];
      Whichpair(site,1) = d_config.whichpair[MID*2*nsites+site*2+1];

      for ( int j=0; j<nneibs; j++ )
        Neiblist(site,j) = d_sqlat.neiblist[j+nneibs*site];
    }

  const int nearnp_size = 4 * nsites;
  for ( int i = threadIdx.x;  i < nearnp_size;  i += block_size )
    Nearnp(i) = d_sqlat.nearnp[i];

  const bool favor_neighbors = dc.neifrac > 0;

  __syncthreads();
  
  // Equilibrate the system by performing a large number of Monte Carlo sweeps
  for ( int i2 = 0; i2 < navesweeps + neqlsweeps; i2++ )
    {
      for (int i = 0; i < npsweep; i++ )
	{
	  if(tid == 0)
	    {
#             ifdef COLLECT_CYCLE_COUNTS
              timer_start_findpairs = clock64();
#             endif
	      
	      // Pick a pair and a spin at random
              ipair = curand_uniform( &localState ) * dc.npairs_randscaler;
	      ispin = curand( &localState ) % 2;
	      oldsite = PList(ipair,ispin);
	      
	      // Find a site where this electron may be moved
              do {
                const float neiprob =
                  favor_neighbors ? curand_uniform( &localState ) : 0;
                const bool donenei = favor_neighbors && neiprob <= dc.neifrac;
                if ( !donenei )
                  {
                    newsite = curand_uniform( &localState )
                      * dc.nsites_randscaler;
                  }
                else
                  {
                    const int inei = curand_uniform( &localState ) 
                      * dc.nneibs_randscaler;
                    newsite = Neiblist(oldsite,inei);
                  }
              }
              while ( dlatocc(newsite) == ispin );

              // Note BT should not occur.
              assert( dlatocc(newsite) != BT );
              spin_flip = dlatocc(newsite) != HL;

	    }

	  __syncthreads();

#         ifdef COLLECT_CYCLE_COUNTS
          if ( !tid )
            {
              timer_start_move_eval = clock64();
              time_findpairs += timer_start_move_eval - timer_start_findpairs;
            }
#         endif

          // Note: these were complex doubles in original C++ code
          double dpbd;
	  
	  if ( spin_flip )
	    {
              const int jpair = Whichpair(newsite,1-ispin);
              dpbd = cpu_d_UpdateConfig_d
                ( ipair, ispin, newsite, jpair, oldsite);
	    }
	  else
	    {
	      dpbd = cpu_DetPByDet
                ( ipair, ispin, newsite, pairfunc, d_config);
	    }
	  
          bool accept;
	  
          if ( !tid )
            {
              const float norm2 = dpbd * dpbd;
              const float prob = curand_uniform( &localState );
              accept = norm2 >= prob;
#             ifdef COLLECT_CYCLE_COUNTS
              timer_start_update = clock64();
              time_move_eval += timer_start_update - timer_start_move_eval;
#             endif
            }
          __syncthreads();
	  
	  if ( accept )
	    {
	      // Accept the move
	      if ( !tid ) d_mcarlo.naccept[MID] += 1;
	      if ( spin_flip )
		{
                  const int jpair = Whichpair(newsite,1-ispin);
                  const double dpbd1 =
                    cpu_DetPByDet
                    ( ipair, ispin, newsite, pairfunc, d_config);
                  // Update the configuration		
                  cpu_UpdateConfig
                    ( ipair, ispin, newsite,
                      pairfunc, dpbd1, d_config);
                  const double dpbd2 = cpu_DetPByDet
                    ( jpair, 1-ispin, oldsite, pairfunc, d_config);
		  
                  // Now, move the other spin electron to the old site
		  
		  cpu_UpdateConfig
                    ( jpair, 1-ispin, oldsite, pairfunc, dpbd2, d_config);

		  if(tid == 0)
		    d_mcarlo.nflip[MID] += 1;
		}
	      else
		{
		  cpu_UpdateConfig
                    ( ipair, ispin, newsite, pairfunc, dpbd, d_config);
		  if(tid == 0)
		    d_mcarlo.nsingle[MID] += 1;
		}
#             ifdef COLLECT_CYCLE_COUNTS
              if ( !tid ) time_update += clock64() - timer_start_update;
#             endif
	    }
	}
      
      if ( i2 < neqlsweeps ) continue;
      
      __syncthreads();
      
      // Now, calculate the energy of the configuration
#     ifdef COLLECT_CYCLE_COUNTS
      if ( !tid ) timer_start_eoc = clock64();
#     endif
      const double energyvalue = cpu_EnergyOfCurrentConfiguration();
      
#     ifdef COLLECT_CYCLE_COUNTS
      if ( !tid )
        {
          const clock_t now = clock64();
          time_eoc += now - timer_start_eoc;
          time_energy_exchange += now - timer_start_energy_exchange;
        }
#     endif
      if ( !tid )
	{
	  idat++;
	  eneloc += energyvalue;
	  
          if ( idat == ndat )
            {
              d_enearr[MID*nbins+nbsamp] = eneloc / double( ndat );
              nbsamp++; eneloc = 0.0; idat = 0;
            }
        }
    }
  
  // Return data.

  if ( tid < npairs )
    {
#if 0
      d_config.plist[MID*2*npairs+(tid)*2] = PList(tid,0);
      d_config.plist[MID*2*npairs+(tid)*2+1] = PList(tid,1);
#endif
      for ( int row=0; row<npairs; row++ )
        for ( int col=0; col<npairs; col++ )
        d_config.PsiInv[MID*npairs*npairs+npairs*row+col] = PsiInv(row,col);
    }
  
  if ( !tid )
    {
      d_mcarlo.nsteps[MID] = nsweeps * npsweep;
      
#     ifdef COLLECT_CYCLE_COUNTS
      clock_t time_total = clock64() - timer_start_thread;
      timers[blockIdx.x].total = time_total;
      timers[blockIdx.x].findpairs = time_findpairs;
      timers[blockIdx.x].move_eval = time_move_eval;
      timers[blockIdx.x].update = time_update;
      timers[blockIdx.x].eoc = time_eoc;
      timers[blockIdx.x].energy_exchange_term_cnt =
        time_energy_exchange_term_cnt;
      timers[blockIdx.x].energy_exchange = time_energy_exchange;
      timers[blockIdx.x].detpbydet[0] = time_detpbydet[0];
      timers[blockIdx.x].detpbydet[1] = time_detpbydet[1];
      timers[blockIdx.x].detpbydet_cnt = time_detpbydet_cnt;
      timers[blockIdx.x].update_config[0] = time_update_config[0];     
      timers[blockIdx.x].update_config[1] = time_update_config[1];
      timers[blockIdx.x].update_config_cnt = time_update_config_cnt;
      timers[blockIdx.x].d_update_d[0] = time_d_update_d[0];     
      timers[blockIdx.x].d_update_d[1] = time_d_update_d[1];
      timers[blockIdx.x].d_update_dw[0] = time_d_update_dw[0];     
      timers[blockIdx.x].d_update_dw[1] = time_d_update_dw[1];
      timers[blockIdx.x].d_update_d_cnt = time_d_update_d_cnt;
      timers[blockIdx.x].d_update_dw_cnt = time_d_update_dw_cnt;
#     endif
    }

  delete s_psi_cache;
}
