#ifndef CUDAMAIN_CUH
#define CUDAMAIN_CUH

#ifdef MAIN
#define Extern
#else
#define Extern extern
#endif

// Declarations used by both CUDA C code and host code.

#define COLLECT_CYCLE_COUNTS

// Define markers to be placed in code next to variables which need to
// be compile-time constants. The source code here assigns a
// placeholder value to these variables (1). The preprocessed code
// will be read by the cactus executable and, using the markers, the
// placeholder values will be replaced by actual values. The
// preprocessed file will then be compiled.
//
#define REPL_TARGET_PREFIX int QQQ_start_marker_
#define REPL_TARGET_MARKER(f) REPL_TARGET_MARKER_1(f)
#define REPL_TARGET_MARKER_1(f) int QQQ_start_marker_##f=0;
#define REPL_TARGET_END int QQQ_end_marker = 0;
#define REPL_TARGET_END_W(f) int QQQ_end_marker_##f = 0;

#define REPL_KEY_MAIN main
#define REPL_TARGET_MAIN REPL_TARGET_MARKER(REPL_KEY_MAIN)

#define REPL_KEY_SITE_IDX site_idx
#define REPL_TARGET_SITE_IDX REPL_TARGET_MARKER(REPL_KEY_SITE_IDX)

#define WP_LG 5
#define WP_SZ (1 << WP_LG)
#define WP_MA (WP_SZ - 1 )

struct Kernel_Timers
{
  // These time by section of code.
  clock_t total, findpairs, move_eval, update, eoc;
  clock_t energy_exchange;

  // These time individual routines.
  clock_t detpbydet[2], update_config[2];
  clock_t d_update_d[2], d_update_dw[2];

  int d_update_d_cnt, d_update_dw_cnt;
  int detpbydet_cnt, update_config_cnt, findpairs_cnt;
  int energy_exchange_term_cnt;
};

Extern __constant__ Kernel_Timers *timers;

// Define structures used to point to data on the GPU
// Define structure to hold particle configuration data for all Markov chains

Extern __constant__ int nsweeps, neqlsweeps, nfac, nbins, naveseeds,
  iseedseed, nattemptmax, navesweeps, npsweep, ndat, nelecs, nearnsets;
Extern __constant__ double INVTOL, selfrac, Jex, Jij;

struct Device_Constants
{
  float neifrac;
  float nsites_randscaler;
  float npairs_randscaler;
  float nneibs_randscaler;
};

Extern __constant__ Device_Constants dc;

typedef struct Pairfunction 
{
  double *pairfn;
  double *Tpairfn;
} Pairfunction;


typedef struct Configuration
{
        int *plist; int *latocc; int *whichpair;
        double *PsiInv;
        double *p_PsiInv; // Padded version.
        bool *modified; double *energy;
        double *kinenergy; double *othenergy;
        double *ext_storage; // Energy exchange term.
} Configuration;

Extern Configuration c_config;

// Define structure to hold Monte Carlo data for all Markov chains
typedef struct MonteCarlo
{
	int *nsteps; int *naccept; int *nsingle; 
	int *nflip; int *nneistep;
} MonteCarlo;

// Define structure to hold lattice data common to all Markov chains
typedef struct Lattice
{
	int *ix; int *iy; double *kx; double *ky;
	int *neiblist; int *ishell; double *thop;
	int *nearnp;
        double *thop_ishell;
} Lattice;

Extern MonteCarlo c_mcarlo;
Extern Lattice c_sqlat;
Extern Pairfunction c_pairfunc;
Extern double *c_enearr;

void cpu_set_vars
(montecarlo& mp, sqlat &latobj, phyparams& phyobj, int psiinv_stride);
extern "C" void* cpu_sysAve(void *id);

#endif
