#include <curand_kernel.h>
#include <assert.h>

// Main file for CUDA C code.

#include "vmc.h"
#include "cudaMain.cuh"

// The declarations in the second line below are changed at run time
// to their correct values (based on the input file, device
// capabilities, etc.). The current values are chosen only to avoid
// compiler errors. The declarations must all be placed on one line.
REPL_TARGET_MAIN
const int npairs = 42; const int nsites = 42; const int nneibs = 42; const int block_size = 512; const bool favor_neighbors = false; const bool psiinv_shared = true; const int thd_per_col = 4;  const int psiinv_stride = 42;  const bool asm_use_nc __attribute__ ((unused)) = false;  const bool use_row_cache = false; const int row_block_size = 32;  const int tune_degree __attribute__ ((unused)) = 8; const int tune_flags __attribute__ ((unused)) = 0;  const bool opt_energy_one_pass = true;

__shared__ clock_t time_update_config[2], time_detpbydet[2];
__shared__ clock_t time_d_update_d[2], time_d_update_dw[2];
__shared__ int time_update_config_cnt, time_detpbydet_cnt;
__shared__ int time_d_update_d_cnt, time_d_update_dw_cnt;
__shared__ int time_energy_exchange_term_cnt;
__shared__ clock_t timer_start_energy_exchange, time_energy_exchange;


const int num_row_blocks = ( npairs + row_block_size - 1 ) / row_block_size;

// Use a single shared memory buffer for DetPByDet and UpdateConfig
#define DBL_BUFFER_SIZE ( block_size + npairs )

REPL_TARGET_SITE_IDX
typedef unsigned char Site_Idx;

const int psiinv_sdim = psiinv_shared ? npairs : 1;

__shared__ Site_Idx s_plist[2][npairs];
__shared__ char s_latocc[nsites];
__shared__ Site_Idx s_whichpair[2*nsites];
__shared__ Site_Idx s_neiblist[nneibs*nsites];
__shared__ double s_psiinv[psiinv_sdim][psiinv_sdim];
__shared__ Site_Idx s_nearnp[4 * nsites];

const int warp_count = block_size >> WP_LG;

// For entire rows or columns of phi
__shared__ double s_psi_cache[use_row_cache ? 1 : warp_count][npairs];

// For caching portions of columns of phi.
__shared__ double sb_psi_cache
[use_row_cache ? warp_count : 1][use_row_cache ? row_block_size : 1];

// For caching multiple rows of psi_inv.
__shared__ double sb_psiinv_cache
[use_row_cache ? row_block_size : 1][use_row_cache ? npairs : 1];

// Storage used for various purposes, including by reduction routine.
volatile __shared__ double dbl_buffer[DBL_BUFFER_SIZE];

#define Nearnp(idx) s_nearnp[idx]

#define PList(pair,spin) s_plist[spin][pair]
#define dlatocc(site) s_latocc[site]
#define Whichpair(site,spin) s_whichpair[(site)*2+(spin)]
#define Neiblist(site,jn) s_neiblist[nneibs*(site)+(jn)]
#define PsiInv(r,c)                                                           \
  ( psiinv_shared                                                             \
    ? s_psiinv[r][c]                                                          \
    : d_config.p_PsiInv[blockIdx.x*npairs*psiinv_stride+psiinv_stride*(r)+(c)] )
#define Pairfunc(r,c) ptx_load(&pairfunc.pairfn[(r)*nsites+(c)])
#define TPairfunc(r,c) ptx_load(&pairfunc.Tpairfn[(c)*nsites+(r)])

__device__ inline double ptx_load(void *addr)
{
#ifdef DC_MAKE_CUP
  if ( !asm_use_nc )
#endif
    return *(double*) addr;
#ifdef DC_MAKE_CUP
  double rv;
  asm("ld.global.nc.f64 %0, [%1];" : "=d"(rv) : "l"(addr) );
  return rv;
#endif
}

#define PsiEltCol(c) s_psi_cache[0][c]
#define PsiEltRow(r) s_psi_cache[0][r]
#define wPsiEltCol(c) s_psi_cache[wid][c]
#define wPsiEltRow(r) s_psi_cache[wid][r]
#define bPsiInv(lr,col) sb_psiinv_cache[lr][col]
#define wbPsiEltRow(lr) sb_psi_cache[wid][lr]

#include "initRand.cu"
#include "k_equil_accum.cu"
#include "sysStat.cu"
