#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstdlib>

namespace miscutils 
{
	void InitializeRandomNumGenerator( int iseed )
	{
		srand( iseed );
	}

	int RandomInteger( int N ) 
	{
		return rand()%N;
	}

	double RandomReal( void ) 
	{
		return ( ( double )rand( ) ) / ( ( double )RAND_MAX );
	}

	void InfoPrint( void )
	{
		std::cout << std::right << "---------------------------------------------------------" << std::endl;
	}

	void InfoPrint( std::string infostr ) 
	{
		std::cout << std::right << "---------------------------------------------------------" << std::endl;
		std::cout << std::right << "  >> " << infostr << std::endl;
	}

	void InfoPrint( std::string infostr, double infod ) 
	{
		if ( ( fabs( infod ) > 1.E4 ) || ( fabs( infod ) < 1.E-4 ) ) 
		{
			std::cout << std::right << std::setw( 35 ) << infostr<< "  : " << std::setw(14) \
				<< std::scientific << std::setprecision(6) << infod << std::endl;
				std::cout << std::resetiosflags( std::ios_base::scientific );
		} 
		else 
		{
			std::cout << std::right << std::setw( 35 ) << infostr << "  : " << std::setw( 12 ) \
				<< std::setprecision( 6 ) << infod << std::endl;
		}
	}

	void InfoPrint( std::string infostr, int infoi ) 
	{
		std::cout << std::right << std::setw( 35 ) << infostr << "  : " << std::setw( 10 ) << infoi << std::endl;
	}

	void InfoPrint( std::string infostr, char infoc )
	{
		std::cout << std::right << std::setw( 35 ) << infostr << "  : " << std::setw( 1 ) << infoc << std::endl;
	}

	void CalculateAverages( int n, double *earr, double& mean, double& sd ) 
	{
		int i;

		if ( n == 0 ) 
		{
			InfoPrint( "Cannot calculate averages" );
			exit( 0 );
		}

		mean = 0.0;
		sd = 0.0;
		for ( i = 0; i < n; i++ ) 
		{
			mean += earr[i];
			sd += earr[i] * earr[i];
		}

		mean /= ( double )n;
		sd /= ( double )n;

		sd = sqrt( ( sd - mean * mean ) / ( double )( n - 1 ) );
	}

}
