
#include "vmc.h"
#include "config.hpp"
#include <fstream>
#include <iomanip>

void
montecarlo::init( sqlat &lobj )
{
	tilt = lobj;

	naccept = 0;
	nsingle = 0;
	nflip = 0;
	nsteps = 0;
	nneistep = 0;
	WriteHistory = false;

	ConfigFile cfg( "Params.cfg" );
	nsweeps = cfg.getValueOfKey<int>( "nsweeps" );
	neqlsweeps = cfg.getValueOfKey<int>( "neqlsweeps" );
	nfac = cfg.getValueOfKey<int>( "nfac" );
	nbins = cfg.getValueOfKey<int>( "nbins" );
	neifrac = cfg.getValueOfKey<int>( "neifrac" );
	INVTOL = cfg.getValueOfKey<double>( "INVTOL" );
	naveseeds = cfg.getValueOfKey<int>( "naveseeds" );
	selfrac = cfg.getValueOfKey<double>( "selfrac" );
	iseedseed = cfg.getValueOfKey<int>( "iseedseed" );
	nMarkov_requested = cfg.getValueOfKey<int>( "nMarkov" );
        nMarkov = 0; // Will be set by main code.

        tune_degree = cfg.getValueOfKey<int>("tune-degree",0);
        tune_flags = cfg.getValueOfKey<int>("tune-flags",0);
        //  printf("Tuning : degree %d  flags 0x%x\n", tune_degree, tune_flags);

	npsweep = nfac * tilt.nelecs;
	navesweeps = nsweeps - neqlsweeps;

	if ( navesweeps % nbins != 0 ) 
	{
		miscutils::InfoPrint( "***ERROR: navesweeps not divisible by nbins", nbins );
		exit( 0 );
	}
	ndat = navesweeps / nbins;

        
	miscutils::InfoPrint( "Monte Carlo Parameters" );
	miscutils::InfoPrint( "Number of Markov chains", nMarkov );
	miscutils::InfoPrint( "Number of sweeps", nsweeps );
	miscutils::InfoPrint( "Number of equilibration sweeps", neqlsweeps );
	miscutils::InfoPrint( "Sweep factor", nfac );
	miscutils::InfoPrint( "Neibor fraction in Monte Carlo step", neifrac );
	miscutils::InfoPrint( "Number of seeds to average over", naveseeds );
	miscutils::InfoPrint( "Selection fraction", selfrac );

	current.CreateConfiguration( tilt.nsites, tilt.npairs );
	nearnsets = 2*current.nsites; 
}
